package graph_algorithms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

public class MetaModuleFinder {

	/* MEMBERS */

	/*
	 * We use CLICK model to create a log-likelihood matrix. We repeat the CLICK
	 * process twice. The first is on the global similarity matrix and requires
	 * an EM step (not done here) to find the mates and non mates distributions.
	 * The second matrix is created by analyzing the results of running our 
	 * differential correlation (DC) analysis on
	 * the real data and comparing it to DC analysis on shuffled data sets.
	 */
	double[][] consistentCorrelationsGraph;
	public double[][] DC_graph;

	float likelihoodThresholdForDC = 0;
	float likelihoodThresholdForCG = 0;
	double sizesConsistencyThr = 0.25;

	// this array holds the singletons
	// if a gene is a singleton than it was not used
	private boolean[] wasUsed;

	/* Other Auxiliary MEMBERS */
	public ArrayList<String> geneNames;
	int minClusterSize = 15;
	
	/* These objects shall contain the final results */
	public LinkedList<HashSet<Integer>> modules;
	public boolean[][] areModulesFriends;
	public LinkedList<MetaModule> meta_modules;

	/**
	 * Constructor
	 * @param c - the consistent correlations graph
	 * @param dc - the DC analysis results
	 * @param ids - the gene names
	 */
	public MetaModuleFinder(double[][] c, double[][] dc,
			ArrayList<String> ids) {
		DC_graph = dc;
		consistentCorrelationsGraph = c;
		geneNames = ids;

	}

	/**
	 * This methods receives as input two distinct sets.
	 * The set elements represent the index of a gene in the matrices.
	 * The algorithm greedily tries to remove a gene from T or S 
	 * until the remaining sets satisfy the meta-module definition.
	 * @param T
	 * @param S
	 * @return
	 */
	public boolean greedyRemoveNodesFromTwoSides(Collection<Integer> T,
			Collection<Integer> S) {
		
		HashSet<Integer> toRemoveS = new HashSet<Integer>();
		for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
			Integer inS = (Integer) iterator.next();
			int currDegree = 0;
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2.hasNext();) {
				Integer inT = (Integer) iterator2.next();
				if (DC_graph[inS][inT] > 0){
					currDegree++;
				}
			}
			if (currDegree <=1){
				toRemoveS.add(inS);
			}
		}
//		
//		for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
//			Integer inS = (Integer) iterator.next();
//			int currDegree = 0;
//			for (Iterator<Integer> iterator2 = S.iterator(); iterator2.hasNext();) {
//				Integer inT = (Integer) iterator2.next();
//				if (inS==inT){
//					continue;
//				}
//				if (clickGlobalmatrix[inS][inT] > 0){
//					currDegree++;
//				}
//			}
//			if (currDegree <=2){
//				toRemoveS.add(inS);
//			}
//		}
//		
		HashSet<Integer> toRemoveT = new HashSet<Integer>();
		for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
			Integer inT = (Integer) iterator.next();
			int currDegree = 0;
			for (Iterator<Integer> iterator2 = S.iterator(); iterator2.hasNext();) {
				Integer inS = (Integer) iterator2.next();
				if (DC_graph[inS][inT] > 0){
					currDegree++;
				}
			}
			if (currDegree <=1){
				toRemoveT.add(inT);
			}
		}
//		
//		for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
//			Integer inT = (Integer) iterator.next();
//			int currDegree = 0;
//			for (Iterator<Integer> iterator2 = T.iterator(); iterator2.hasNext();) {
//				Integer inS = (Integer) iterator2.next();
//				if (inS==inT){
//					continue;
//				}
//				if (clickGlobalmatrix[inS][inT] > 0){
//					currDegree++;
//				}
//			}
//			if (currDegree <=2){
//				toRemoveT.add(inT);
//			}
//		}
		S.removeAll(toRemoveS);
		T.removeAll(toRemoveT);
		
//		System.out.println("cleaning stage removed: " + toRemoveS.size()+toRemoveT.size());
		
		Hashtable<Integer, Double> inTNode2WithinScore = getWithinSetScores(T);
		Hashtable<Integer, Double> inSNode2WithinScore = getWithinSetScores(S);

		Hashtable<Integer, Double> fromTtoSScores = new Hashtable<Integer, Double>();
		Hashtable<Integer, Double> fromStoTScores = new Hashtable<Integer, Double>();

		for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
			Integer inS = (Integer) iterator.next();
			fromStoTScores.put(inS, 0.0);
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2
					.hasNext();) {
				Integer inT = (Integer) iterator2.next();
				if (!fromTtoSScores.containsKey(inT)) {
					fromTtoSScores.put(inT, 0.0);
				}
				fromStoTScores.put(inS, fromStoTScores.get(inS)
						+ DC_graph[inS][inT]);

				fromTtoSScores.put(inT, fromTtoSScores.get(inT)
						+ DC_graph[inS][inT]);
			}
		}

		while (true) {

			if (T.size() <= 1 || S.size() <= 1) {
				return false;
			}
			

			double minScore = Double.MAX_VALUE;
			int minCombScoreIndex = -1;
			double min1 = minScore;
			double min2 = minScore;
			double minBetween = Double.MAX_VALUE;
			int minBetweenIndex = -1;
			double minWithin = Double.MAX_VALUE;
			int minWithinIndex = -1;
			int minNodeSelected = -1;
			
			for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
				Integer node = (Integer) iterator.next();
				if (inTNode2WithinScore.get(node) + fromTtoSScores.get(node) < minScore) {
					minScore = inTNode2WithinScore.get(node)
							+ fromTtoSScores.get(node);
					min1 = inTNode2WithinScore.get(node);
					min2 = fromTtoSScores.get(node);
					minCombScoreIndex = node;
				}
				if (inTNode2WithinScore.get(node) < minWithin){
					minWithin = inTNode2WithinScore.get(node);
					minWithinIndex = node;
				}
				if (fromTtoSScores.get(node) < minBetween){
					minBetween = fromTtoSScores.get(node);
					minBetweenIndex = node;
				}
			}

			for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
				Integer node = (Integer) iterator.next();
				if (inSNode2WithinScore.get(node) + fromStoTScores.get(node) < minScore) {
					minScore = inSNode2WithinScore.get(node) + fromStoTScores.get(node);
					min1 = inSNode2WithinScore.get(node);
					min2 = fromStoTScores.get(node);
					minNodeSelected = node;
					minCombScoreIndex = node;
				}
				if (inSNode2WithinScore.get(node) < minWithin){
					minWithin = inSNode2WithinScore.get(node);
					minWithinIndex = node;
				}
				if (fromStoTScores.get(node) < minBetween){
					minBetween = fromStoTScores.get(node);
					minBetweenIndex = node;
				}
			}		
//			if (T.size()==1 && S.size()==1 && minBetween > 0){
//				return true;
//			}
			
			if  (min1 <= likelihoodThresholdForDC
					&& min2 <= likelihoodThresholdForCG){
				minNodeSelected=minCombScoreIndex;
			}
			else if (min1 <= likelihoodThresholdForDC){
				minNodeSelected = minBetweenIndex;
			}
			else{
				minNodeSelected = minWithinIndex;
			}

			if (min1 > likelihoodThresholdForCG
					&& min2 > likelihoodThresholdForDC
						&& minBetween > likelihoodThresholdForDC
							&& minWithin > likelihoodThresholdForCG) {
				return true;
			}

			if (T.contains(minNodeSelected)){
				
				inTNode2WithinScore.remove(minNodeSelected);
				T.remove(minNodeSelected);
				fromTtoSScores.remove(minNodeSelected);
				
				for (Iterator<Integer> iterator2 = T.iterator(); iterator2.hasNext();) {
					Integer inT = (Integer) iterator2.next();
					double score = inTNode2WithinScore.get(inT);
					score -= consistentCorrelationsGraph[inT][minNodeSelected];
					inTNode2WithinScore.put(inT, score);
				}
				
				for (Iterator<Integer> iterator2 = S.iterator(); iterator2.hasNext();) {
					Integer inS = (Integer) iterator2.next();
					double score = fromStoTScores.get(inS);
					score -= DC_graph[inS][minNodeSelected];
					fromStoTScores.put(inS, score);
				}
				
			}
			else if (S.contains(minNodeSelected)){

				inSNode2WithinScore.remove(minNodeSelected);
				S.remove(minNodeSelected);
				fromStoTScores.remove(minNodeSelected);
				
				for (Iterator<Integer> iterator2 = S.iterator(); iterator2.hasNext();) {
					Integer inS = (Integer) iterator2.next();
					double score = inSNode2WithinScore.get(inS);
					score -= consistentCorrelationsGraph[inS][minNodeSelected];
					inSNode2WithinScore.put(inS, score);
				}
				
				for (Iterator<Integer> iterator2 = T.iterator(); iterator2
						.hasNext();) {
					Integer inT = (Integer) iterator2.next();
					double score = fromTtoSScores.get(inT);
					score -= DC_graph[inT][minNodeSelected];
					fromTtoSScores.put(inT, score);
				}
			}
		}
	}

	public Hashtable<Integer, Double> getWithinSetScores(Collection<Integer> T) {
		
		Hashtable<Integer, Double> node2Score = new Hashtable<Integer, Double>();
		for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
			Integer inT = (Integer) iterator.next();
			node2Score.put(inT, 0.0);
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2
					.hasNext();) {
				Integer secondInT = (Integer) iterator2.next();
				if (inT == secondInT){
					continue;
				}
				double score = node2Score.get(inT);
				score += consistentCorrelationsGraph[inT][secondInT];
				node2Score.put(inT, score);
			}
		}
		return node2Score;
	}

	public void findSeedPairs() {

		modules = new LinkedList<HashSet<Integer>>();
		meta_modules = new LinkedList<MetaModule>();

		wasUsed = new boolean[DC_graph.length];
		Arrays.fill(wasUsed, false);

		// get all edges from the DC process
		LinkedList<ComparableEdge> edges = getEdgesForAlgo(likelihoodThresholdForDC);
		double thr = likelihoodThresholdForDC;
		while (edges.size() > 100000) {
			thr++;
			edges = getEdgesForAlgo(thr);
		}

		Hashtable<Integer, HashSet<Integer>> neighborsListsInDCAGraph = createNeighborsListsInGraph();
		sortEdgesByScores(edges, neighborsListsInDCAGraph);

		int itersCounter = 0;
		while (edges.size() > 0 && itersCounter < 10000) {
			itersCounter++;
			
			ComparableEdge currentEdge = edges.removeLast();
			int u = currentEdge.u;
			int v = currentEdge.v;
			if (wasUsed[u] || wasUsed[v]) {
				continue;
			}

			// get initial neighbors lists
			HashSet<Integer> neighborsOfU = new HashSet<Integer>(neighborsListsInDCAGraph.get(u));
			HashSet<Integer> neighborsOfV = new HashSet<Integer>(neighborsListsInDCAGraph.get(v));
			
			if (neighborsOfU.size() < 1 || neighborsOfV.size() < 1) {
				continue;
			}

//			System.out.print("side1:" + neighborsOfU.size());
//			System.out.print(",side2:" + neighborsOfV.size());

			removeIntersection(neighborsOfU, neighborsOfV);
			
//			System.out.print("\tsizes after removal: " + "side1:"
//					+ neighborsOfU.size() + ",side2:" + neighborsOfV.size());

			if (neighborsOfU.size() < 1 || neighborsOfV.size() < 1) {
//				System.out.println();
				continue;
			}
			

//			ClusterPair currPair = new ClusterPair(new HashSet<Integer>(neighborsOfU),
//					new HashSet<Integer>(neighborsOfV), this); 
//			System.out.print(" Before removal " + neighborsOfU.size() + ","
//					+ neighborsOfV.size() +" score is " + currPair.betweenLLRScore);
			greedyRemoveNodesFromTwoSides(neighborsOfU, neighborsOfV);
//			currPair = new ClusterPair(new HashSet<Integer>(neighborsOfU),
//					new HashSet<Integer>(neighborsOfV), this); 
//			
//			System.out.print(" After removal " + neighborsOfU.size() + ","
//					+ neighborsOfV.size() +" score is " + currPair.betweenLLRScore);
			
			if (neighborsOfU.size() < 1 || neighborsOfV.size() < 1) {
//				System.out.println();
				continue;
			}
			
			double s1 = neighborsOfU.size();
			double s2 = neighborsOfV.size();

			if (s1 >= 2 && s2 >= 2
					&& ((Math.min(s1 / s2, s2 / s1)) >= sizesConsistencyThr)) {

				modules.add(new HashSet<Integer>(neighborsOfV));
				modules.add(new HashSet<Integer>(neighborsOfU));
				MetaModule currPair = new MetaModule(new HashSet<Integer>(neighborsOfU),
						new HashSet<Integer>(neighborsOfV), this); 
				meta_modules.add(currPair);
				
//				System.out.print(" found pair: " + neighborsOfU.size() + ","
//						+ neighborsOfV.size() +" score is " + currPair.betweenLLRScore +"\n");


				// go over all covered genes and mark as used
				for (Iterator<Integer> iterator = neighborsOfU.iterator(); iterator
						.hasNext();) {
					Integer neighbor = (Integer) iterator.next();
					wasUsed[neighbor] = true;
				}
				for (Iterator<Integer> iterator = neighborsOfV.iterator(); iterator
						.hasNext();) {
					Integer neighbor = (Integer) iterator.next();
					wasUsed[neighbor] = true;
				}
				
				// removing u and v from the graph
				removeCurrentListFromGraph(neighborsListsInDCAGraph, neighborsOfU);
				removeCurrentListFromGraph(neighborsListsInDCAGraph, neighborsOfV);
//				neighborsListsInDCAGraph = createNeighborsListsInGraph();

				// update edges lists
				sortEdgesByScores(edges, neighborsListsInDCAGraph);
				
			} else {
//				System.out.println();
			}
		}
	}



	public Hashtable<Integer, HashSet<Integer>> createNeighborsListsInGraph() {
		Hashtable<Integer,HashSet<Integer>> neighborsListsInDCAGraph = new Hashtable<Integer,HashSet<Integer>>();
		for (int i = 0; i < DC_graph.length; i++) {
			neighborsListsInDCAGraph.put(i,new HashSet<Integer>());
		}

		for (int i = 0; i < DC_graph.length; i++) {
			if (wasUsed[i]) {
				continue;
			}
			for (int j = 0; j < i; j++) {
				if (wasUsed[j]) {
					continue;
				}
				if (DC_graph[i][j]>likelihoodThresholdForDC){
					neighborsListsInDCAGraph.get(i).add(j);
					neighborsListsInDCAGraph.get(j).add(i);
				}
			}
		}
		return neighborsListsInDCAGraph;
	}

	public LinkedList<ComparableEdge> getEdgesForAlgo(double likelihoodThresholdForDCA) {
		LinkedList<ComparableEdge> edges = new LinkedList<ComparableEdge>();
		for (int i = 0; i < DC_graph.length; i++) {
			for (int j = 0; j < i; j++) {
				if (DC_graph[i][j] > likelihoodThresholdForDCA) {
					ComparableEdge e = new ComparableEdge(i, j, 0);
					edges.add(e);
				}
			}
		}
		return edges;
	}



	

	public void calculateMetaGraph() {

//		System.out.println("calculating meta graph for " + modules.size()
//				+ " nodes");
		areModulesFriends = new boolean[modules.size()][modules.size()];
		for (int i = 0; i < areModulesFriends.length; i++) {
			HashSet<Integer> firstModule = modules.get(i);
//			double firstSize = firstModule.size();
			for (int j = 0; j < i; j++) {
				HashSet<Integer> secondModule = modules.get(j);
//				double secondSize = secondModule.size();
//				double ratio = Math.min(firstSize / secondSize, secondSize
//						/ firstSize);
//				if (ratio < sizesConsistencyThr) {
//					continue;
//				}
				// get between scores
				double score = 0;
				for (Iterator<Integer> iterator = secondModule.iterator(); iterator
						.hasNext();) {
					Integer node1 = (Integer) iterator.next();

					for (Iterator<Integer> iterator2 = firstModule.iterator(); iterator2
							.hasNext();) {
						Integer node2 = (Integer) iterator2.next();
						score += DC_graph[node1][node2];
					}

				}

				if (score > likelihoodThresholdForDC) {
					areModulesFriends[i][j] = true;
					areModulesFriends[j][i] = true;
				}
			}
		}

	}


	public LinkedList<LinkedList<Integer>> addSingletonsToModules() {

		boolean[] wasUsed = new boolean[DC_graph.length];
		for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator.hasNext();) {
			HashSet<Integer> m = (HashSet<Integer>) iterator.next();
			for (Iterator<Integer> iterator2 = m.iterator(); iterator2.hasNext();) {
				Integer inm = (Integer) iterator2.next();
				wasUsed[inm]=true;
			}
		}
		
		ArrayList<Integer> sigs = new ArrayList<Integer>();
		for (int i = 0; i < wasUsed.length; i++) {
			if (!wasUsed[i]){
				sigs.add(i);
			}
		}
		
		SingletonsAndMetaModulesMerger merger = new SingletonsAndMetaModulesMerger(meta_modules, sigs, this);
		merger.doMerges(1,0);
		meta_modules = new LinkedList<MetaModule>();
		for (int i = 0; i < merger.pairs.length; i++) {
			meta_modules.add(merger.pairs[i]);
		}
		
		modules = merger.getModulesList();
		return merger.getMetaModules();
	}

	public void getFinalSolution(int maxNumClusters, int minSize) {

//		modules = doHierarchicalClustering(0,clickGlobalmatrix);
//		calculateMetaGraph();
//		pairs = new LinkedList<ClusterPair>();
//		for (int i = 0; i < modules.size(); i++) {
//			for (int j = 0; j < i; j++) {
//				if (areModulesFriends[i][j]){
//					ClusterPair pair = new ClusterPair(modules.get(i), modules.get(j), this);
//					System.out.println("current llr score: " + pair.betweenLLRScore + ", size is: " + pair.getUnion().size());
//					pairs.add(pair);
//				}
//			}
//		}
		
		findSeedPairs();
	
		MetaModulesMerger merger = new MetaModulesMerger(meta_modules);
		merger.doMerges();
		meta_modules = new LinkedList<MetaModule>();
		for (int i = 0; i < merger.pairs.length; i++) {
			meta_modules.add(merger.pairs[i]);
		}
		modules = merger.getModulesList();
		addSingletonsToModules();
		
//		merger = new SeedPairsMerger(pairs);
//		merger.doMerges();
//		pairs = new LinkedList<ClusterPair>();
//		for (int i = 0; i < merger.pairs.length; i++) {
//			pairs.add(merger.pairs[i]);
//		}
//		addSingletonsToModules();
//		modules = merger.getModulesList();
		
		Collections.sort(meta_modules);
		System.out.println("******* sorted meta clusters ******");

		modules = new LinkedList<HashSet<Integer>>();
		LinkedList<MetaModule> newPairs = new LinkedList<MetaModule>();
		for (int i = meta_modules.size() - 1; i >= 0; i--) {
			if (newPairs.size() >= maxNumClusters) {
				break;
			}
			System.out.println("current llr score: " + meta_modules.get(i).betweenLLRScore + ", size is: " + meta_modules.get(i).getUnion().size());
			if (meta_modules.get(i).getUnion().size() < minSize) {
				continue;
			}

			newPairs.add(meta_modules.get(i));
			modules.add(meta_modules.get(i).U);
			modules.add(meta_modules.get(i).V);
		}
		
		meta_modules = newPairs;
	}

	public void removeZeroDegreesMetaNodes() {
		LinkedList<HashSet<Integer>> nModules;
		int[] degrees = new int[modules.size()];
		for (int i = 0; i < areModulesFriends.length; i++) {
			for (int j = 0; j < i; j++) {
				if (areModulesFriends[i][j]) {
					degrees[i]++;
					degrees[j]++;
				}
			}
		}

		nModules = new LinkedList<HashSet<Integer>>();
		for (int i = 0; i < degrees.length; i++) {
			if (degrees[i] > 0) {
				nModules.add(modules.get(i));
			}
		}
		modules = nModules;
		calculateMetaGraph();
	}

	public void setLikelihoodThresholdForDC(float likelihoodThresholdForDCA) {
		this.likelihoodThresholdForDC = likelihoodThresholdForDCA;
	}
	
	
	/**
	 * Static auxiliary methods
	 * 
	 */
	
	public static void sortEdgesByScores(LinkedList<ComparableEdge> edges,
			Hashtable<Integer,HashSet<Integer>> neighborsListsInDCAGraph) {
		LinkedList<ComparableEdge> toRemove = new LinkedList<ComparableEdge>();
		for (Iterator<ComparableEdge> iterator = edges.iterator(); iterator.hasNext();) {
			ComparableEdge e = (ComparableEdge) iterator.next();
			int u = e.u;
			int v = e.v;
			double deg1 = neighborsListsInDCAGraph.get(u).size();
			double deg2 = neighborsListsInDCAGraph.get(v).size();
			e.score = deg1 + deg2;
			if (e.score==0){
				toRemove.add(e);
			}
			else{
				e.score *= Math.min(deg1 / deg2, deg2 / deg1);
			}
		}
		
		edges.removeAll(toRemove);
		Collections.sort(edges);
	}

	public static void removeCurrentListFromGraph
	(Hashtable<Integer,HashSet<Integer>> neighborsListsInDCAGraph, HashSet<Integer> set) {
		
		for (Iterator<Integer> iterator2 = set.iterator(); iterator2.hasNext();) {
			Integer inSet = (Integer) iterator2.next();
			neighborsListsInDCAGraph.put(inSet, new HashSet<Integer>());
		}
		
		for (Iterator<Integer> iterator = neighborsListsInDCAGraph.keySet().iterator(); iterator.hasNext();) {
			Integer index = (Integer) iterator.next();
			HashSet<Integer> list =neighborsListsInDCAGraph.get(index);
			if (list.size()==0){
				continue;
			}
			list.removeAll(set);
		}
	}


	public static void removeIntersection(HashSet<Integer> neighborsOfU,
			HashSet<Integer> neighborsOfV) {
		
		LinkedList<Integer> intersection = new LinkedList<Integer>();
		for (Iterator<Integer> iterator = neighborsOfU.iterator(); iterator
				.hasNext();) {
			Integer neighborOfU = (Integer) iterator.next();
			if (neighborsOfV.contains(neighborOfU)) {
				intersection.add(neighborOfU);
			}
		}
		neighborsOfU.removeAll(intersection);
		neighborsOfV.removeAll(intersection);
	}
}
