package main_scripts;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import math_utils.ExpressionMatrix;

import EM_LLR_Utils.LLRScoresCalculator;
import graph_algorithms.HierarchicalClusterer;
import graph_algorithms.MetaModule;
import graph_algorithms.MetaModuleFinder;

/**
 * 
 * This class runs the DICER algorithm and outputs the analysis results.
 * 
 * * For running the main (i.e. the runnable jar), the format of the input
 * should be (in this order): 1. A gene expression matrix file. This is a tab
 * delimited file in which the first row contains the condition names, and the
 * first column contains the gene names. For example, see the LungCancer.txt
 * file LungCancer.zip. 2. A "classes file". This file contains partitioning of
 * the conditions in the matrix to groups. We assume that conditions from the
 * same groups are adjacent in the matrix. This file is tab delimited, in each
 * row there are three columns: the start index of the current group
 * (enumeration of the conditions starts from zero),the end index, and the group
 * name. For example, see the classesFile.txt in LungCancer.zip 3. A number
 * between 0 and the number of classes -1 that specifies the class of interest.
 * 4. A name of an output file.
 * 
 * OPTIONAL: 5. 1 if you want to print the DC and CG graphs (as matrices).
 * 
 * All output files prefixes are the name of the output path specified by the
 * user. It creates the following files, according to the suffixes: 1.
 * _Modules.txt - a flat file with the gene to modules assignments. in this
 * file, up or down specifies from which DICER analysis the modules where
 * detected. 2. _metaGraph.sif - a SIF file that represents the DC map among the
 * modules 3. _metaGraph.txt - a file with the edge scores for _metaGraph.sif 4.
 * The final output file is the output path indicated by the user and it
 * contains all gene to meta-module or DC cluster mappings.
 * 
 */
public class DICER_main {

	/* Members */
	// Inputs from the user (should come from the args array)
	String gePath;
	String classesPath;
	int classIndex;
	String outPath;

	// Memebers for the statistical class-specific analysis
	protected TscoresCalculator tScoresCalc;
	protected float[][] realTscoresGraph;
	protected float[] randomScores;
	protected LLRScoresCalculator llrAnalysisForTscores;
	protected LLRScoresCalculator globalCorrelationsLLRAnalyzer;

	// The results of the statistical analysis: up DC, down DC and the CG graphs
	protected double[][] dGGraph;
	protected double[][] uGGraph;
	protected double[][] consistentCorrelationsGraph;

	// The objects used for clustering and module extraction
	protected MetaModuleFinder downClusterer;
	protected MetaModuleFinder upClusterer;
	protected MetaModuleFinder upMetaModulesFinder;
	protected MetaModuleFinder downMetaModulesFinder;

	// Objects used for keeping the results
	protected LinkedList<HashSet<Integer>> downCorrelatedClusters;
	protected LinkedList<HashSet<Integer>> upCorrelatedClusters;
	protected Hashtable<String, HashSet<String>> modules = new Hashtable<String, HashSet<String>>();
	protected Hashtable<String, HashSet<String>> metaClusters = new Hashtable<String, HashSet<String>>();
	private double K=2;

	/**
	 * Constructors
	 */

	/**
	 * The format of the input should be (in this order): 1. A gene expression
	 * matrix file. This is a tab delimited file in which the first row contains
	 * the condition names, and the first column contains the gene names. For
	 * example, see the LungCancer.txt file LungCancer.zip. 2. A "classes file".
	 * This file contains partitioning of the conditions in the matrix to
	 * groups. We assume that conditions from the same groups are adjacent in
	 * the matrix. This file is tab delimited, in each row there are three
	 * columns: the start index of the current group (enumeration of the
	 * conditions starts from zero),the end index, and the group name. For
	 * example, see the classesFile.txt in LungCancer.zip 3. A number between 0
	 * and the number of classes -1 that specifies the class of interest. 4. A
	 * name of an output file.
	 */
	public DICER_main(String[] args) {
		gePath = args[0];
		classesPath = args[1];
		classIndex = Integer.parseInt(args[2]);
		outPath = args[3];
		tScoresCalc = new TscoresCalculator(gePath, classesPath, "pearson");
	}

	public DICER_main(String gP, String cP, int cla, String out) {
		gePath = gP;
		classesPath = cP;
		classIndex = cla;
		outPath = out;
		tScoresCalc = new TscoresCalculator(gePath, classesPath, "pearson");
	}

	public DICER_main(ExpressionMatrix mat, String cP, int cla) {

		classesPath = cP;
		classIndex = cla;
		tScoresCalc = new TscoresCalculator(mat, cP, "pearson");
	}

	/**
	 * This method runs the algorithm and keeps all statistical and clustering
	 * results in the fitting modules.
	 * 
	 * @param doStatistical
	 *            : a boolean. Specifies if to perform the statistical analysis
	 *            or to read the graphs instead
	 * @param printMatrices
	 *            : a boolean. Specifies if to write the graphs to files or not
	 */
	public void runAlgo(boolean doStatistical, boolean printMatrices) {

		if (doStatistical) {
			doStatisticalAnalysis();
		} else {
			String path = (outPath.split("\\."))[0];
			readStatisticalMatrices(path);
		}

		if (printMatrices) {
			String path = (outPath.split("\\."))[0];
			printStatisticalAnalysisMatrices(path);
		}

		downClusterer = new MetaModuleFinder(null, dGGraph,
				tScoresCalc.data.getGenesID());
		downCorrelatedClusters = HierarchicalClusterer
				.doHierarchicalClustering(2, dGGraph, 15);

		upClusterer = new MetaModuleFinder(null, uGGraph,
				tScoresCalc.data.getGenesID());
		upCorrelatedClusters = HierarchicalClusterer.doHierarchicalClustering(
				2, uGGraph, 15);

		System.out.println("num down clusters: "
				+ downCorrelatedClusters.size());
		System.out.println("num up clusters " + upCorrelatedClusters.size());

		upMetaModulesFinder = new MetaModuleFinder(consistentCorrelationsGraph,
				uGGraph, tScoresCalc.data.getGenesID());
		upMetaModulesFinder.getFinalSolution(10, 20);

		downMetaModulesFinder = new MetaModuleFinder(
				consistentCorrelationsGraph, dGGraph,
				tScoresCalc.data.getGenesID());
		downMetaModulesFinder.getFinalSolution(10, 20);

		getMetaModules("upMM", upMetaModulesFinder, metaClusters);
		getMetaModules("downMM", downMetaModulesFinder, metaClusters);

		getModules(downMetaModulesFinder, "Module_down_analysis", modules);
		getModules(upMetaModulesFinder, "Module_up_analysis", modules);

	}

	/**
	 * This method performs statistical operations: 1. Compare random and real
	 * T-scores. 2. Calculate the consistent correlations graph. Progress
	 * remarks are printed to the standard output.
	 */
	public void doStatisticalAnalysis() {
		DCGraphsCalculator calculator = new DCGraphsCalculator(tScoresCalc,
				classIndex);
		calculator.setK(K);
		calculator.doStatisticalAnalysis();
		realTscoresGraph = calculator.realTscoresGraph;
		randomScores = calculator.randomScores;
		llrAnalysisForTscores = calculator.llrAnalysisForTscores;
		globalCorrelationsLLRAnalyzer = calculator.globalCorrelationsLLRAnalyzer;
		dGGraph = calculator.dGGraph;
		uGGraph = calculator.uGGraph;
		consistentCorrelationsGraph = calculator.consistentCorrelationsGraph;
	}

	/**
	 * Set of IO methods
	 */

	/**
	 * This method prints DICER flow results. All output files prefixes are the
	 * name of the output path specified by the user. It creates the following
	 * files, according to the suffixes: 1. _Modules.txt - a flat file with the
	 * gene to modules assignments. in this file, up or down specifies from
	 * which DICER analysis the modules where detected. 2. _metaGraph.sif - a
	 * SIF file that represents the DC map among the modules 3. _metaGraph.txt -
	 * a file with the edge scores for _metaGraph.sif 4. The final output file
	 * is the output path indicated by the user and it contains all gene to
	 * meta-module or DC cluster mappings.
	 */
	public void printAnalysisResults() {
		// print results

		String otherPathsPrefix = outPath;

		try {
			String seedsPath = otherPathsPrefix + "_AllModules.txt";
			BufferedWriter writerForSeeds = new BufferedWriter(new FileWriter(
					seedsPath));
			printModules(upMetaModulesFinder, "up", writerForSeeds, 0, modules);
			printModules(downMetaModulesFinder, "down", writerForSeeds, 0,
					modules);

			String largeSeedsPath = otherPathsPrefix + "_Modules.txt";
			BufferedWriter writerForLargeSeeds = new BufferedWriter(
					new FileWriter(largeSeedsPath));
			printModules(upMetaModulesFinder, "up", writerForLargeSeeds, 10,
					null);
			printModules(downMetaModulesFinder, "down", writerForLargeSeeds,
					10, null);

			String sifPath = otherPathsPrefix + "_metaGraph.sif";
			String scoresPath = otherPathsPrefix + "_metaGraph.txt";
			BufferedWriter writerForSif = new BufferedWriter(new FileWriter(
					sifPath));
			BufferedWriter writerForTxt = new BufferedWriter(new FileWriter(
					scoresPath));
			printMetaGraph(upMetaModulesFinder, downMetaModulesFinder,
					writerForSif, writerForTxt);

			BufferedWriter writer = new BufferedWriter(new FileWriter(outPath));
			printMetaClusters("downMM", writer, downMetaModulesFinder,
					metaClusters);
			printMetaClusters("upMM", writer, upMetaModulesFinder, metaClusters);
			printClusters(downCorrelatedClusters, "downCluster",
					tScoresCalc.data.genesID, writer);
			printClusters(upCorrelatedClusters, "upCluster",
					tScoresCalc.data.genesID, writer);

			writer.flush();
			writer.close();
			writerForSeeds.flush();
			writerForSeeds.close();
			writerForSif.flush();
			writerForSif.close();
			writerForTxt.flush();
			writerForTxt.close();
			writerForLargeSeeds.flush();
			writerForLargeSeeds.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// perform enrichment analyses
		// String keggPath = "humanGenes2KeggNames.txt";
		// String complexesPath = "Entrez2ProteinComplexes.txt";
		// String AnatomicalAnnot = "EntrezId2AnatomicalData.txt";
		// System.out.println("****** seed enrichment results ******");
		// System.out.println("###KEGG###");
		// HyperGeometricEnrichmentAnalysis anal = new
		// HyperGeometricEnrichmentAnalysis(
		// keggPath, comparator.data.getGenesID(), seeds);
		// anal.printResults(0.05);
		// System.out.println("###Complexes###");
		// anal = new HyperGeometricEnrichmentAnalysis(complexesPath,
		// comparator.data.getGenesID(), seeds);
		// anal.printResults(0.1);
		// System.out.println("###Anatomical###");
		// anal = new HyperGeometricEnrichmentAnalysis(AnatomicalAnnot,
		// comparator.data.getGenesID(), seeds);
		// anal.printResults(0.05);
		//
		// System.out.println("****** meta modules enrichment results ******");
		// System.out.println("###KEGG###");
		// anal = new HyperGeometricEnrichmentAnalysis(keggPath,
		// comparator.data.getGenesID(), metaClusters);
		// anal.printResults(0.05);
		// System.out.println("###Complexes###");
		// anal = new HyperGeometricEnrichmentAnalysis(complexesPath,
		// comparator.data.getGenesID(), metaClusters);
		// anal.printResults(0.1);
		// System.out.println("###Anatomical###");
		// anal = new HyperGeometricEnrichmentAnalysis(AnatomicalAnnot,
		// comparator.data.getGenesID(), metaClusters);
		// anal.printResults(0.05);

		// perform fold change tests
		// if (seeds.size()>0){
		// System.out.println("analyzing meta modules fold change");
		// DifferentialClusteringSolutionSignificanceEstimator.calculateDiffStatisticsForPairsForOurAlgo(comparator,
		// seeds);
		// }
		// if (downCorrelatedModules.size()>0){
		// System.out.println("analyzing down correlated clusters fold change");
		// DifferentialClusteringSolutionSignificanceEstimator.calculateDiffStatistics(comparator,
		// downCorrelatedModules);
		// }
		// if (upCorrelatedModules.size()>0){
		// System.out.println("analyzing up correlated clusters fold change");
		// DifferentialClusteringSolutionSignificanceEstimator.calculateDiffStatistics(comparator,
		// upCorrelatedModules);
		// }

	}

	public static float[] floatListToArray(LinkedList<Float> positiveScores) {
		float[] posScores = new float[positiveScores.size()];
		int counter = 0;
		for (Iterator<Float> iterator = positiveScores.iterator(); iterator
				.hasNext();) {
			Float score = (Float) iterator.next();
			posScores[counter] = score;
			counter++;
		}
		return posScores;
	}

	protected static void printModules(MetaModuleFinder clust, String type,
			BufferedWriter writer, int minSize,
			Hashtable<String, HashSet<String>> seeds) {
		int counter = 0;
		for (Iterator<HashSet<Integer>> iterator = clust.modules.iterator(); iterator
				.hasNext();) {
			HashSet<Integer> module = (HashSet<Integer>) iterator.next();
			if (module.size() < minSize) {
				continue;
			}
			for (Iterator<Integer> iterator2 = module.iterator(); iterator2
					.hasNext();) {
				Integer geneIndex = (Integer) iterator2.next();
				String geneName = clust.geneNames.get(geneIndex);
				String seedName = type + "_" + counter;
				if (minSize <= 1) {
					if (!seeds.containsKey(seedName)) {
						HashSet<String> set = new HashSet<String>();
						set.add(geneName);
						// seeds.put(seedName, set);
					} else {
						seeds.get(seedName).add(geneName);
					}
				}

				try {
					writer.write(geneName + "\t" + seedName + "\n");
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

			counter++;
		}
	}

	protected static void getModules(MetaModuleFinder clust, String type,
			Hashtable<String, HashSet<String>> seeds) {
		int counter = 0;
		for (Iterator<HashSet<Integer>> iterator = clust.modules.iterator(); iterator
				.hasNext();) {
			HashSet<Integer> module = (HashSet<Integer>) iterator.next();
			if (module.size() < 2) {
				continue;
			}
			for (Iterator<Integer> iterator2 = module.iterator(); iterator2
					.hasNext();) {
				Integer geneIndex = (Integer) iterator2.next();
				String geneName = clust.geneNames.get(geneIndex);
				String seedName = type + "_" + counter;

				if (!seeds.containsKey(seedName)) {
					HashSet<String> set = new HashSet<String>();
					set.add(geneName);
					seeds.put(seedName, set);
				} else {
					seeds.get(seedName).add(geneName);
				}

			}

			counter++;
		}
	}

	protected static void printMetaClusters(String type, BufferedWriter writer,
			MetaModuleFinder clusterer, Hashtable<String, HashSet<String>> mms) {
		int countModules = -1;

		for (Iterator<MetaModule> iterator = clusterer.meta_modules.iterator(); iterator
				.hasNext();) {
			MetaModule metaCluster = (MetaModule) iterator.next();
			countModules++;

			for (Iterator<Integer> iterator2 = metaCluster.getUnion()
					.iterator(); iterator2.hasNext();) {
				Integer index = (Integer) iterator2.next();
				String gene = clusterer.geneNames.get(index);
				String clusterName = type + "_" + countModules + "_" + "_"
						+ metaCluster.betweenLLRScore;
				if (!mms.containsKey(clusterName)) {
					HashSet<String> mm = new HashSet<String>();
					mm.add(gene);
					// mms.put(clusterName, mm);
				} else {
					mms.get(clusterName).add(gene);
				}
				try {

					writer.write(gene + "\t" + clusterName + "\n");
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

		}

	}

	protected static void getMetaModules(String type,
			MetaModuleFinder clusterer, Hashtable<String, HashSet<String>> mms) {
		int countModules = -1;

		for (Iterator<MetaModule> iterator = clusterer.meta_modules.iterator(); iterator
				.hasNext();) {
			MetaModule metaCluster = (MetaModule) iterator.next();
			countModules++;

			for (Iterator<Integer> iterator2 = metaCluster.getUnion()
					.iterator(); iterator2.hasNext();) {
				Integer index = (Integer) iterator2.next();
				String gene = clusterer.geneNames.get(index);
				String clusterName = type + "_" + countModules + "_" + "_"
						+ metaCluster.betweenLLRScore;
				if (!mms.containsKey(clusterName)) {
					HashSet<String> mm = new HashSet<String>();
					mm.add(gene);
					mms.put(clusterName, mm);
				} else {
					mms.get(clusterName).add(gene);
				}
			}

		}

	}

	protected static int printClusters(
			LinkedList<HashSet<Integer>> downCluster, String type,
			ArrayList<String> geneNames, BufferedWriter writer) {
		int countModules = -1;
		for (Iterator<HashSet<Integer>> iterator = downCluster.iterator(); iterator
				.hasNext();) {
			HashSet<Integer> cluster = (HashSet<Integer>) iterator.next();

			countModules++;
			for (Iterator<Integer> iterator2 = cluster.iterator(); iterator2
					.hasNext();) {
				Integer gene = (Integer) iterator2.next();
				try {
					writer.write(geneNames.get(gene) + "\t" + type + "_"
							+ countModules + "\n");

				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

		}
		return countModules;
	}

	protected static void printMetaGraph(MetaModuleFinder clustererUp,
			MetaModuleFinder clustererDown, BufferedWriter writer1,
			BufferedWriter writer2) {

		LinkedList<HashSet<Integer>> allSeeds = new LinkedList<HashSet<Integer>>();
		allSeeds.addAll(clustererDown.modules);
		allSeeds.addAll(clustererUp.modules);

		try {
			writer2.write("\n");
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		for (int i = 0; i < allSeeds.size(); i++) {
			HashSet<Integer> module1 = allSeeds.get(i);
			String name1 = "down_" + i;
			if (i >= clustererDown.modules.size()) {
				name1 = "up_" + (i - clustererDown.modules.size());
			}
			for (int j = 0; j < i; j++) {
				HashSet<Integer> module2 = allSeeds.get(j);
				String name2 = "down_" + j;
				if (j >= clustererDown.modules.size()) {
					name2 = "up_" + (j - clustererDown.modules.size());
				}
				double upScore = 0;
				double downScore = 0;
				for (Iterator<Integer> iterator = module2.iterator(); iterator
						.hasNext();) {
					Integer inM2 = (Integer) iterator.next();

					for (Iterator<Integer> iterator2 = module1.iterator(); iterator2
							.hasNext();) {
						Integer inM1 = (Integer) iterator2.next();
						if (inM1 == inM2) {
							continue;
						}
						upScore += clustererUp.DC_graph[inM1][inM2];
						downScore += clustererDown.DC_graph[inM1][inM2];
					}
				}

				try {
					if (upScore > 0) {
						writer1.write(name1 + " EdgeUp " + name2 + "\n");
						writer2.write(name1 + " (EdgeUp) " + name2 + " = "
								+ upScore + "\n");
					}
					if (downScore > 0) {
						writer1.write(name1 + " EdgeDown " + name2 + "\n");
						writer2.write(name1 + " (EdgeDown) " + name2 + " = "
								+ -downScore + "\n");
					}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}
		}

	}

	public boolean[][] getMetaGraph() {

		LinkedList<HashSet<Integer>> allSeeds = new LinkedList<HashSet<Integer>>();
		MetaModuleFinder clustererUp = upMetaModulesFinder;
		MetaModuleFinder clustererDown = downMetaModulesFinder;
		allSeeds.addAll(clustererDown.modules);
		allSeeds.addAll(clustererUp.modules);

		boolean[][] graph = new boolean[tScoresCalc.data.matrix.length][tScoresCalc.data.matrix.length];

		for (int i = 0; i < allSeeds.size(); i++) {
			HashSet<Integer> module1 = allSeeds.get(i);
			for (int j = 0; j < i; j++) {
				HashSet<Integer> module2 = allSeeds.get(j);
				double upScore = 0;
				double downScore = 0;
				for (Iterator<Integer> iterator = module2.iterator(); iterator
						.hasNext();) {
					Integer inM2 = (Integer) iterator.next();

					for (Iterator<Integer> iterator2 = module1.iterator(); iterator2
							.hasNext();) {
						Integer inM1 = (Integer) iterator2.next();
						if (inM1 == inM2) {
							continue;
						}
						upScore += clustererUp.DC_graph[inM1][inM2];
						downScore += clustererDown.DC_graph[inM1][inM2];
					}
				}

				if (upScore > 0 || downScore > 0) {
					for (Iterator<Integer> iterator = module2.iterator(); iterator
							.hasNext();) {
						Integer inM2 = (Integer) iterator.next();

						for (Iterator<Integer> iterator2 = module1.iterator(); iterator2
								.hasNext();) {
							Integer inM1 = (Integer) iterator2.next();
							if (inM1 == inM2) {
								continue;
							}
							graph[inM1][inM2]=true;
						}
					}

				}
			}
		}
		return graph;

	}
	
	public boolean[][] getMetaGraph(LinkedList<HashSet<Integer>> allSeeds) {

		MetaModuleFinder clustererUp = upMetaModulesFinder;
		MetaModuleFinder clustererDown = downMetaModulesFinder;

		boolean[][] graph = new boolean[tScoresCalc.data.matrix.length][tScoresCalc.data.matrix.length];

		for (int i = 0; i < allSeeds.size(); i++) {
			HashSet<Integer> module1 = allSeeds.get(i);
			for (int j = 0; j < i; j++) {
				HashSet<Integer> module2 = allSeeds.get(j);
				double upScore = 0;
				double downScore = 0;
				for (Iterator<Integer> iterator = module2.iterator(); iterator
						.hasNext();) {
					Integer inM2 = (Integer) iterator.next();

					for (Iterator<Integer> iterator2 = module1.iterator(); iterator2
							.hasNext();) {
						Integer inM1 = (Integer) iterator2.next();
						if (inM1 == inM2) {
							continue;
						}
						upScore += clustererUp.DC_graph[inM1][inM2];
						downScore += clustererDown.DC_graph[inM1][inM2];
					}
				}

				if (upScore > 0 || downScore > 0) {
					for (Iterator<Integer> iterator = module2.iterator(); iterator
							.hasNext();) {
						Integer inM2 = (Integer) iterator.next();

						for (Iterator<Integer> iterator2 = module1.iterator(); iterator2
								.hasNext();) {
							Integer inM1 = (Integer) iterator2.next();
							if (inM1 == inM2) {
								continue;
							}
							graph[inM1][inM2]=true;
						}
					}

				}
			}
		}
		return graph;

	}

	protected void printStatisticalAnalysisMatrices(String pathName) {
		printMatrix(dGGraph, pathName + "_Gdown.txt");
		printMatrix(uGGraph, pathName + "_Gup.txt");
		printMatrix(consistentCorrelationsGraph, pathName + "_GC.txt");
	}

	protected void readStatisticalMatrices(String pathName) {
		dGGraph = readMatrix(pathName + "_Gdown.txt");
		uGGraph = readMatrix(pathName + "_Gup.txt");
		consistentCorrelationsGraph = readMatrix(pathName + "_GC.txt");
	}

	public static void printMatrix(double[][] mat, String path) {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(path));
			for (int i = 0; i < mat.length; i++) {
				for (int j = 0; j < mat.length; j++) {
					if (j == 0) {
						writer.write("" + mat[i][j]);
					} else {
						writer.write("\t" + mat[i][j]);
					}

				}
				writer.write("\n");
			}
			writer.flush();
			writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static double[][] readMatrix(String path) {
		double[][] mat = null;
		try {
			BufferedReader reader = new BufferedReader(new FileReader(path));
			String line = reader.readLine();
			line.trim();
			String[] firstLine = line.split("\t");
			int n = firstLine.length;
			mat = new double[n][n];
			for (int i = 0; i < firstLine.length; i++) {
				mat[0][i] = Double.parseDouble(firstLine[i]);
			}

			// read next lines
			for (int i = 1; i < mat.length; i++) {
				line = reader.readLine();
				line.trim();
				firstLine = line.split("\t");
				for (int j = 0; j < firstLine.length; j++) {
					mat[i][j] = Double.parseDouble(firstLine[j]);
				}
			}
			reader.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return mat;
	}

	public static void main(String[] args) {
		if (args.length < 4) {
			System.out
					.println("usage: <GE matrix> <classes file><classIndex><file for results>");
			System.out
					.println("Optional: <out=1 if you want to print the CG and DC graphs>");
			System.out
			.println("Optional: <K=value - default value is 2, for larger modules we recommend using 1>");
			System.out
					.println(" All output files prefixes are the name of the output path specified by the user.\n"
							+ "It creates the following files, according to the suffixes:\n"
							+ "\t\t1. _Modules.txt - a flat file with the gene to modules assignments.\n"
							+ "\t\t  in this file, up or down specifies from which DICER analysis the modules where detected.\n"
							+ "\t\t2. _metaGraph.sif -  a SIF file that represents the DC map among the modules.\n"
							+ "\t\t3. _metaGraph.txt - a file with the edge scores for _metaGraph.sif\n"
							+ "\t\t4. The final output file is the output path indicated by the user and it contains\n"
							+ "\t\tall gene to meta-module or DC cluster mappings.");
			System.exit(1);
		}

		DICER_main analyzer = new DICER_main(args);
		
		double k=2;
		if (args.length>4 && args[4].startsWith("K")){
			k=Double.parseDouble(args[4].split("=")[1]);
		}
		if (args.length>5 && args[5].startsWith("K")){
			k=Double.parseDouble(args[4].split("=")[1]);
		}
		analyzer.K=k;
		
		if (args.length > 4 && args[4].equalsIgnoreCase("out=1")) {
			analyzer.runAlgo(true, true);
		} 
		else if(args.length > 5 && args[5].equalsIgnoreCase("out=1")){
			analyzer.runAlgo(true, true);
		}
		else {
			analyzer.runAlgo(true, false);
		}

		analyzer.printAnalysisResults();

	}

	public Hashtable<String, HashSet<String>> getMetaClusters() {
		return metaClusters;
	}

	public LinkedList<HashSet<Integer>> getUpCorrelatedClusters() {
		return upCorrelatedClusters;
	}

	public MetaModuleFinder getDownClusterer() {
		return downClusterer;
	}

	public Hashtable<String, HashSet<String>> getModules() {
		return modules;
	}

}
